Like many other API providers, it’s also possible to generate API Tokens to be used (in HTTP Header) for authentication. An API token ties to a Admin with equal or less privileges.

Notes:

1. an API token generated for a specific admin has the same privilege as the user
2. an API token will be automatically removed if not used for > 90 days
    

SSO admins cannot generate these API tokens. Refer Org level API tokens (/api/v1/docs/Org#api-token) which can have privileges of a specific Org/Site for more information.

# Create API Token

Note that the key is only available during creation time.

```
POST /api/v1/self/apitokens

```

### Response

```json
{
    "id": "7cc917df-630a-1508-1889-e18028052dde",
    "key": "MEaT4pEYG9hgHSUzddRUlT0TDRp3quf"
}

```

# Use API Token

To use API token, add a Authorization header when making an API request like the following:

```
Authorization: Token 

GET /api/v1/self

```

# List Current API Tokens

```
GET /api/v1/self/apitokens

```

### Response

```json
[
    {
        "id": "7cc917df-630a-1508-1889-e18028052dde",
        "key": "MEaT...3quf"
    }
]

```

# Delete an API Token

```
DELETE /api/v1/self/apitokens/:apitoken_id

```

# Change Admin information, password, and enable/disable two-factor auth

```
PUT /api/v1/self

```

### Request

```json
{
    "password": "foryoureyesonly",
    "first_name": "John",
    "last_name": "Smith",
    "phone": "14081112222",
    "phone2": "14083334444",
    "persona": "security",
    "enable_two_factor": true
}

```

### Meanings

| Name | Type | Description |
| --- | --- | --- |
| `password` | `string` | new password |
| `first_name` | `string` | first name of logged-in user |
| `last_name` | `string` | last name of logged-in user |
| `phone` | `string` | phone number (numbers only, including country code) |
| `phone2` | `string` | secondary phone number |
| `enable_two_factor` | `boolean` | to enable or disable two-factor authentication |
| `two_factor_verified` | `boolean` | if enable_two_factor=true, whether it’s verified. |

- `enable_two_factor`. Note that when two-factor authentication is enabled (from disabled-state), the `two_factor_seed` is re-generated and `two_factor_verified` is reset to false. The OTP will need to be verified before it takes effect.
    

# Change Email

We require the user to verify that they actually own the email address they intend to change it to.

```
POST /api/v1/self/update

```

### Request

```json
{
    "email": "new@mistsys.com"
}

```

### Meanings

| Name | Type | Description |
| --- | --- | --- |
| `email` | `string` | new email address |

### Response if OK

```
Status: 200 OK

```

After the API call, the user will receive an email to the new email address with a link like https://manage.mist.com/verify/update?expire=:exp_time&email=:admin_email&token=:token

Upon clicking the link, the user is provided with a login page to authenticate using existing credentials. After successful login, the email address of the user gets updated

Note: The request parameter `email` can be used by UI to validate that the current session (if any) belongs to the admin or provide a login page (by pre-populating the email on login screen). UI can also use the request parameter `expire` to validate token expiry.

### Response if invalid email address

```
Status: 400 Bad Request

```

### Response if new email address already exists

```json
Status: 400 Bad Request

{
    "detail": "email already existed"
}

```

# Verify Email change

```
POST /api/v1/self/update/verify/:token

```

### Response if OK

With correct verification, the email address of the user will be updated

```
Status: 200 OK

```

### Response if invalid or expired token

```json
Status: 400 Bad Request

{
    "detail": "invalid token"
}

```

### Response if request comes from authenticated users other than admin

```json
Status: 400 Bad Request

{
    "detail": "invalid token"
}

```

### Response if new email address already exists

```json
Status: 400 Bad Request

{
    "detail": "email already existed"
}

```

# Generate QR code for verification

```
GET /api/v1/self/two_factor/token?by=qrcode

```

### Request Parameters

| Name | Type | Description |
| --- | --- | --- |
| `by` | `string` | qrcode |

# Verify Two-factor (OTP)

To verify two-factor authentication by using a code generated by app (e.g. Google Authenticator, Authy) or by SMS. Upon successful verification, the `otp_verified` will be set to true if it hasn’t already been.

```
POST /api/v1/self/two_factor/verify

```

### Request

```json
{
    "two_factor": "123456"
}

```

### Response when 2FA verification is successful

```
Status: 200 OK

```
